from tabulate import tabulate

from omg.common.helper import age

# Special function to output node
# Generate output table if -o not set or 'wide'
# We will create an array of array and then print if with tabulate
def node_out(t, ns, res, output, show_type):
    output_nodes=[]
    # header
    # we will append the header array at last after sorting
    if output == 'wide':
        header = ['NAME','STATUS','ROLES','AGE','VERSION',
            'INTERNAL-IP','EXTERNAL-IP','OS-IMAGE','KERNEL-VERSION','CONTAINER-RUNTIME']
    else:
        header = ['NAME','STATUS','ROLES','AGE','VERSION']
    # pods
    for node in res:
        n = node['res']
        row = []
        # name
        if show_type:
            row.append(t + '/' + n['metadata']['name'])
        else:
            row.append(n['metadata']['name'])
        # status
        ready = next( c['status'] for c in n['status']['conditions'] if c['type'] == 'Ready')
        if ready == 'True':
            status = "Ready"
        else:
            status = "NotReady"
        if 'unschedulable' in n['spec'] and n['spec']['unschedulable']:
            status += ",SchedulingDisabled"
        row.append(status)
        # roles
        roles = []
        for label in n['metadata']['labels']:
            if label.startswith('node-role.kubernetes.io'):
                roles.append(label.split('/')[1])
        row.append(','.join(roles))
        # age
        ct = str(n['metadata']['creationTimestamp'])
        ts = node['gen_ts']
        row.append(age(ct,ts))
        # version
        ver = n['status']['nodeInfo']['kubeletVersion']
        row.append(ver)
        if output == "wide":
            # internal/external ip
            i_ip = '<none>'
            e_ip = '<none>'
            addresses = n['status']['addresses']
            for a in addresses:
                if a['type'] == 'InternalIP':
                    i_ip = a['address']
                if a['type'] == 'ExternalIP':
                    e_ip = a['address']
            row.append(i_ip)
            row.append(e_ip)
            # os-image
            osimg = n['status']['nodeInfo']['osImage']
            row.append(osimg)
            # kernelVersion
            kver = n['status']['nodeInfo']['kernelVersion']
            row.append(kver)
            crt = n['status']['nodeInfo']['containerRuntimeVersion']
            row.append(crt)
        output_nodes.append(row)

    # sort by 2nd column (role)
    sorted_output = sorted(output_nodes, key=lambda x: x[2])
    sorted_output.insert(0,header)

    print(tabulate(sorted_output,tablefmt="plain"))



from tabulate import tabulate

from omg.common.helper import age


def service_out(t, ns, res, output, show_type):
    output_res=[[]]
    # header
    if ns == '_all':
        output_res[0].append('NAMESPACE')
    output_res[0].extend(['NAME','TYPE','CLUSTER-IP','EXTERNAL-IP','PORT(S)','AGE'])
    if output == 'wide':
        output_res[0].extend(['SELECTOR'])
    # resources
    for r in res:
        svc = r['res']
        row = []
        # namespace (for --all-namespaces)
        if ns == '_all':
            row.append(svc['metadata']['namespace'])
        # name
        if show_type:
            row.append(t + '/' + svc['metadata']['name'])
        else:
            row.append(svc['metadata']['name'])
        # type
        row.append(svc['spec']['type'])
        # cluster-ip
        if 'clusterIP' in svc['spec']:
            row.append(svc['spec']['clusterIP'])
        else:
            row.append('<none>')
        # external-ip
        if 'externalName' in svc['spec']:
            row.append(svc['spec']['externalName'])
        elif 'externalIP' in svc['spec']:
            row.append(svc['spec']['externalIP'])
        else:
            row.append('<none>')
        # ports
        if 'ports' in svc['spec']:
            ports = ','.join(
                str(p['port']) + '/' + p['protocol'] 
                for p in svc['spec']['ports'] if 'port' in p
            )
            row.append(ports)
        else:
            row.append('<none>')
        # age
        try:
            ct = str(svc['metadata']['creationTimestamp'])
            ts = r['gen_ts']
            row.append(age(ct,ts))
        except:
            row.append('Unknown')
        #selector (-o wide)
        if output == 'wide':
            if 'selector' in svc['spec']:
                selector = ",".join(
                    [ label+'='+value for label,value in svc['spec']['selector'].items() ]
                )
                row.append(selector)
            else:
                row.append('<none>')

        output_res.append(row)

    print(tabulate(output_res,tablefmt="plain"))



from tabulate import tabulate

from omg.common.helper import age

# Special function to output pod
# Generate output table if -o not set or 'wide'
# We will create an array of array and then print if with tabulate
def pod_out(t, ns, res, output, show_type):
    output_pods=[[]]
    # header
    if ns == '_all':
        output_pods[0].append('NAMESPACE')
    if output == 'wide':
        output_pods[0].extend(['NAME','READY','STATUS','RESTARTS','AGE','IP','NODE'])
    else:
        output_pods[0].extend(['NAME','READY','STATUS','RESTARTS','AGE'])
    # pods
    for pod in res:
        p = pod['res']
        row = []
        # namespace (for --all-namespaces)
        if ns == '_all':
            row.append(p['metadata']['namespace'])
        # name
        if show_type:
            row.append(t + '/' + p['metadata']['name'])
        else:
            row.append(p['metadata']['name'])
        # containers/ready count
        containers = str(len(p['spec']['containers']))
        containers_ready = str(len([ r for r in p['status']['containerStatuses']  if r['ready'] == True ]))
        row.append(containers_ready+'/'+containers)
        # status
        row.append(p['status']['phase'])
        # restarts
        row.append(max([ r['restartCount'] for r in p['status']['containerStatuses'] ]))
        # age
        pod_ct = str(p['metadata']['creationTimestamp'])
        gen_ts = pod['gen_ts']
        row.append(age(pod_ct,gen_ts))
        # pod ip and node (if -o wide)
        if output == "wide":
            if 'podIP' in p['status']:
                row.append(p['status']['podIP'])
            else:
                row.append('')
            if 'nodeName' in p['spec']:
                row.append(p['spec']['nodeName'])
            else:
                row.append('')

        output_pods.append(row)

    print(tabulate(output_pods,tablefmt="plain"))



from tabulate import tabulate

from omg.common.helper import age


# endpoint out
def ep_out(t, ns, res, output, show_type):
    output_res=[[]]
    # header
    if ns == '_all':
        output_res[0].append('NAMESPACE')
    output_res[0].extend(['NAME','ENDPOINTS','AGE'])
    # resources
    for r in res:
        ep = r['res']
        row = []
        # namespace (for --all-namespaces)
        if ns == '_all':
            row.append(ep['metadata']['namespace'])
        # name
        if show_type:
            row.append(t + '/' + ep['metadata']['name'])
        else:
            row.append(ep['metadata']['name'])
        # endpoints
        endpoints = []
        if 'subsets' in ep:
            for s in ep['subsets']:
                if 'addresses' in s and 'ports' in s:
                    for a in s['addresses']:
                        for p in s['ports']:
                            endpoints.append(
                                str(a['ip']) + ':' + str(p['port'])
                            )
        if len(endpoints) < 4:
            row.append(
                ','.join(endpoints)
            )
        else:
            row.append(
                ','.join(endpoints[:3]) +
                ' + ' +
                str(len(endpoints)-3) +
                ' more...'
            )
            row.append('<none>')
        # age
        try:
            ct = str(ep['metadata']['creationTimestamp'])
            ts = r['gen_ts']
            row.append(age(ct,ts))
        except:
            row.append('Unknown')

        output_res.append(row)

    print(tabulate(output_res,tablefmt="plain"))



import sys, os
from omg.common.config import Config

def use(a):
    c = Config(fail_if_no_path=False)
    p = a.mg_path
    # We traverse up to 3 levels to find the must-gather
    # At each leve if it has only one dir and we check inside it
    # When we see see the dir /namespaces and /cluster-scoped-resources, we assume it
    for i in [1,2,3]:
        if os.path.isdir(p):
            if ( os.path.isdir( os.path.join(p, 'namespaces')) and
                 os.path.isdir( os.path.join(p, 'cluster-scoped-resources')) ):
                full_path = os.path.abspath(p)
                c.save(path=full_path)
                print('Using: ',p)
                break
            elif len(os.listdir(p)) == 1:
                p = os.path.join(p,os.listdir(p)[0])
            else:
                print('[ERROR] Invalid must-gather path. Please point to the extracted must-gather directory')
                break
        else:
            print('[ERROR] Invalid path. Please give path to the extracted must-gather')
            break
        



import sys, yaml, json
from omg.common.config import Config
from omg.common.resource_map import map_res

# The high level function that gets called for any "get" command
# This fucntion Processes/indentifies the objects, they can be in various formats e.g:
#   get pod httpd
#   get pods httpd1 httpd2
#   get dc/httpd pod/httpd1
#   get routes
#   get pod,svc
# We get all these args (space separated) in the array a.objects
# We'll process them and normalize them in a python dict (objects)
# Once we have one of more object to get,
# and call the respective get function
# (this function is looked up from common/resource_map)
def get_main(a):
    # a = args passed from cli
    # Check if -A/--all-namespaces is set
    # else, Set the namespace
    # -n/--namespace takes precedence over current project 
    if a.all_namespaces is True:
        ns = '_all'
    else:
        if a.namespace is not None:
            ns = a.namespace
        elif Config().project is not None:
            ns = Config().project
        else:
            ns = None

    # We collect the resources types/names in this dict
    # e.g for `get pod httpd1 httpd2` this will look like:
    #   objects = { 'pod': ['httpd1', 'httpd2'] }
    # e.g, for `get pod,svc` this will look like:
    #   objects = { 'pod': ['_all'], 'service': ['_all'] }
    objects = {}


    last_object = []
    for o in a.objects:
        # Case where we have a '/'
        # e.g omg get pod/httpd
        if '/' in o:
            if not last_object:
                pre = o.split('/')[0]
                r_type = map_res(pre)['type']
                r_name = o.split('/')[1]
                # If its a valid resource type, apppend it to objects
                if r_type is not None:
                    if r_type in objects:
                        objects[r_type].append(r_name)
                    else:
                        objects[r_type] = [r_name]
                else:
                    print("[ERROR] Invalid object type: ",pre)
                    sys.exit(1)
            else:
                # last_object was set, meaning this should be object name
                print("[ERROR] There is no need to specify a resource type as a separate argument when passing arguments in resource/name form")
                sys.exit(1)

        # Case where we have a ',' e.g `get dc,svc,pod httpd`
        # These all will be resource_types, not names,
        # resource_name will come it next iteration (if any)
        elif ',' in o:
            if not last_object:
                r_types = o.split(',')
                for rt in r_types:
                    check_rt = map_res(rt)
                    if check_rt is None:
                        print("[ERROR] Invalid object type: ",rt)
                        sys.exit(1)
                    else:
                        last_object.append(check_rt['type'])
            else:
                # last_object was set, meaning this should be object name
                print("[ERROR] Invalid resources to get: ", a.objects)
                sys.exit(1)

        # Simple word (without , or /)
        # If last_object was not set, means this is a resource_type
        elif not last_object:
            check_rt = map_res(o)
            if check_rt is not None:
                last_object = [ check_rt['type'] ]
            else:
                print("[ERROR] Invalid resource type to get: ", o)
        # Simple word (without , or /)
        # If the last_object was set, means we got resource_type last time,
        # and this should be a resource_name. 
        elif last_object:
            for rt in last_object:
                if rt in objects:
                    objects[rt].append(o)
                else:
                    objects[rt] = [o]
            last_object = []
        else:
            # Should never happen
            print("[ERROR] Invalid resources to get: ", o)
            sys.exit(1)
    # If after going through all the args, we have last_object set
    # means we didn't get a resource_name for these resource_type.
    # i.e, we need to get all names
    if last_object:
        for rt in last_object:
            check_rt = map_res(rt)
            objects[check_rt['type']] = ['_all']

    # Debug
    # print(objects)

    # Object based routing
    # i.e, call the get function for all the requested types
    # then call the output function or simply print if its yaml/json
    for rt in objects.keys():
        rt_info = map_res(rt)
        get_func = rt_info['get_func']
        getout_func = rt_info['getout_func']
        yaml_loc = rt_info['yaml_loc']
        need_ns = rt_info['need_ns']

        # Call the get function to get the resoruces
        res = get_func(rt, ns, objects[rt], yaml_loc, need_ns)

        # Error out if no objects/resources were collected
        if len(res) == 0:
            print('No resources found for type "%s" found in namespace "%s" '%(rt,ns))
        # Yaml dump if -o yaml
        elif a.output == 'yaml':
            if len(res) == 1:
                print(yaml.dump(res[0]['res']))
            elif len(res) > 1:
                print(yaml.dump({'apiVersion':'v1','items':[cp['res']for cp in res]}))
        # Json dump if -o json
        elif a.output == 'json':
            if len(res) == 1:
                print(json.dumps(res[0]['res'],indent=4))
            elif len(res) > 1:
                print(json.dumps({'apiVersion':'v1','items':[cp['res']for cp in res]},indent=4))
        # Call the respective output fucntion if -o is not set or -o wide
        elif a.output in [None, 'wide']:
            # If we displaying more than one resource_type,
            # we need to display resource_type with the name (type/name)
            if len(objects) > 1:
                show_type = True
            else:
                show_type = False
            getout_func(rt, ns, res, a.output, show_type)



from tabulate import tabulate

from omg.common.helper import age


# Simple out put with just name and age
def ev_out(t, ns, res, output, show_type):
    output_res=[[]]
    # header
    if ns == '_all':
        output_res[0].append('NAMESPACE')
    output_res[0].extend(['LAST SEEN','TYPE','REASON','OBJECT','MESSAGE'])
    # resources
    for r in res:
        ev = r['res']
        row = []
        # namespace (for --all-namespaces)
        if ns == '_all':
            row.append(ev['metadata']['namespace'])

        # last
        last_ts = str(ev['lastTimestamp'])
        gen_ts = r['gen_ts']
        row.append(age(last_ts,gen_ts))
        # type
        row.append(ev['type'])
        # reason
        row.append(ev['reason'])
        # object
        if 'involvedObject' in ev:
            inv_obj = ev['involvedObject']
            row.append(
                inv_obj['kind'].lower() + '/' + inv_obj['name']
            )
        else:
            row.append('<none>')
        # message
        row.append(ev['message'])

        output_res.append(row)

    print(tabulate(output_res,tablefmt="plain"))



import sys, os, yaml

from omg.common.config import Config

# This function finds the respective yamls and returns the resouces that match
# args = resource_type (e.g pod), namespace, resource_names (e.g, httpd)
def from_yaml(rt, ns, names, yaml_loc, need_ns):
    mg_path = Config().path
    yaml_path = os.path.join(mg_path, yaml_loc)
    if need_ns:
        # Error out if it needs ns and its not set.
        if ns is None:
            print("[ERROR] Namespace not set. Select a project (omg project) or specify a namespace (-n)")
            sys.exit(1)
        # Get all namespace names if we need all
        elif ns == '_all':
            nses = os.listdir( os.path.join(mg_path, 'namespaces') )
            yaml_paths = [ yaml_path%(n) for n in nses if os.path.isfile(yaml_path%(n)) ]
        else:
            yaml_paths = [ yaml_path%(ns) ] if os.path.isfile(yaml_path%(ns)) else []
    else:
        # if yaml_loc in resource_map is a dir, we will read all yamls from this dir
        if os.path.isdir(yaml_path):
            yaml_paths = [ os.path.join(yaml_path, y) for y in os.listdir(yaml_path) if y.endswith('.yaml') ]
        else:
            yaml_paths = [ yaml_path ] if os.path.isfile(yaml_path) else []

    #Debug
    # print(yaml_paths)
    
    # Collect the resources
    collected=[]
    for yp in yaml_paths:
        try:
            # record when was this yaml generated (to calc age)
            gen_ts = os.path.getmtime(yp)
            with open(yp, 'rb') as yf:
                try:
                    res = yaml.safe_load(yf)
                except:
                    print("[ERROR] Parsing error in ", yp)
                    sys.exit(1)
        except:
            print("[ERROR] Could not read file:", yp)
            sys.exit(1)

        # add objects to collected if name matches
        # or if we want to get all the objects (e.g `get pods`)
        if 'items' in res:
            # we got a list
            if res['items'] is not None and len(res['items']) > 0:
                collected.extend(
                    [ {'res':r,'gen_ts':gen_ts}
                        for r in res['items']
                            if r['metadata']['name'] in names or '_all' in names ]
                )
            # else the list was empty/none, we dont' add anything to collected
        elif 'metadata' in res:
            # we got a single item
            collected.extend(
                [ {'res':res,'gen_ts':gen_ts} ] 
                if res['metadata']['name'] in names or '_all' in names else []
            )

    return collected


def from_yaml_dir(t, ns, names):
    pass


import sys, os
from omg.common.config import Config


def project(a):
    c = Config()
    ns_dir = os.path.join(c.path,'namespaces')
    if a is None or a.project is None:
        # print current project
        if c.project is None:
            print('No project selected')
        else:
            print('Using project "%s" on must-gather "%s"' % (c.project,c.path))
    else:
        # Set current project
        if os.path.isdir(os.path.join(ns_dir, a.project)):
            if a.project == c.project:
                print('Already on project "%s" on server "%s"' % (c.project,c.path))
            else:
                c.save(project=a.project)
                print('Now using project "%s" on must-gather "%s"' % (c.project,c.path))
        else:
            print('[ERROR] Project %s not found in %s'%(a.project,ns_dir))


def projects(a):
    c = Config()
    ns_dir = os.path.join(c.path,'namespaces')
    projects = [ p for p in os.listdir(ns_dir) 
                    if os.path.isdir(os.path.join(ns_dir,p)) ]
    print("You have access to the following projects and can switch between them with 'omg project <projectname>':")
    print()
    for proj in projects:
        if proj == c.project:
            print('  * ',proj)
        else:
            print('    ',proj)
    print()
    project(None)



import sys, os
from dateutil.parser import parse
from dateutil.relativedelta import relativedelta
from datetime import datetime, timedelta

from omg.common.config import Config

# This function is used to calculate age of the objects
# We compare the time reported in the yaml definition,
# with the timestamp of the yaml file (i.e, when it was generated).
# This should give us age of the object at the time of must-gather
# obj_time time is in standard format e.g: '2020-06-04T22:10:41Z'
# file_ts is unix ts e.g: 1590912494.0 (returned by os.path.getmtime)

def age(obj_time, file_ts):

    try:
        dt1 = parse(obj_time, ignoretz=True)
        dt2 = datetime.utcfromtimestamp(file_ts)
        rd = relativedelta(dt2, dt1)
    except:
        return 'Unknown'

    if rd.days > 0:
        return str(rd.days)+'d'
    elif rd.hours > 9:
        return str(rd.hours)+'h'
    elif rd.hours > 0 and rd.hours < 10:
        return str(rd.hours)+'h'+str(rd.minutes)+'m'
    elif rd.minutes > 9:
        return str(rd.minutes) + 'm'
    elif rd.minutes > 0 and rd.minutes < 10:
        return str(rd.minutes)+'m'+str(rd.seconds)+'s'
    else:
        return str(rd.seconds)+'s'


import os, sys

from omg.common.config import Config

def log(a):
    if a.all_namespaces is True:
        print('[ERROR] All Namespaces is not supported with log')
        sys.exit(1)
    else:
        if a.namespace is not None:
            ns = a.namespace
        elif Config().project is not None:
            ns = Config().project
        else:
            # ns not set
            print('[ERROR] Namespaces/project not set')
            sys.exit(1)

    ns_dir = os.path.join(Config().path,'namespaces', ns)

    # check if ns directory exists
    if not os.path.isdir(ns_dir):
        print('[ERROR] Namespace not found:', ns)
        sys.exit(1)
    
    # pod
    if '/' in a.resource:
        r_type = a.resource.split('/')[0]
        pod = a.resource.split('/')[1]
        if r_type not in ['pod','pods']:
            print('[ERROR] Can not print logs of type:',r_type)
            sys.exit(1)
    else:
        pod = a.resource

    # check if top pod directory exits
    pod_dir = os.path.join(ns_dir, 'pods', pod)
    if not os.path.isdir(pod_dir):
        print('[ERROR] Pod directory not found:', pod_dir)
        sys.exit(1)
    
    # Containers are dirs in pod_dir
    containers = [ c for c in os.listdir(pod_dir) 
                    if os.path.isdir(os.path.join(pod_dir,c))]
    # If we have > 1 containers and -c/--container is not speficied, error out
    if len(containers) == 0:
        print('[ERROR] No container directory not found in pod direcotry:', pod_dir)
        sys.exit(1)
    elif len(containers) > 1:
        if a.container is None:
            print('[ERROR] This pod has more than one containers:')
            print('       ', str(containers))
            print('        Use -c/--container to specify the container')
            sys.exit(1)
        else:
            con_to_log = a.container
    else: # len(containers) == 1
        con_to_log = containers[0]

    if a.previous:
        log_f = 'previous.log'
    else:
        log_f = 'current.log'

    file_to_cat = os.path.join(pod_dir,con_to_log,con_to_log,'logs',log_f)
    if not os.path.isfile(file_to_cat):
        print('[ERROR] Log file missing: ', file_to_cat)
        sys.exit(1)
    
    # ouput the log file
    print(file_to_cat)

    with open(file_to_cat, 'r') as ol:
            print(ol.read())



from tabulate import tabulate

from omg.common.helper import age


# Simple out put with just name and age
def rs_out(t, ns, res, output, show_type):
    output_res=[[]]
    # header
    if ns == '_all':
        output_res[0].append('NAMESPACE')
    output_res[0].extend(['NAME','DESIRED','CURRENT','READY','AGE'])
    if output == 'wide':
        output_res[0].extend(['CONTAINERS', 'IMAGES','SELECTOR'])
    # resources
    for r in res:
        rs = r['res']
        row = []
        # namespace (for --all-namespaces)
        if ns == '_all':
            row.append(rs['metadata']['namespace'])
        # name
        if show_type:
            row.append(t + '/' + rs['metadata']['name'])
        else:
            row.append(rs['metadata']['name'])
        # desired, current, ready
        for x in ['replicas', 'availableReplicas', 'readyReplicas']:
            if x in rs['status']:
                row.append(rs['status'][x])
            else:
                row.append('0')
        # age
        try:
            ct = str(rs['metadata']['creationTimestamp'])
            ts = r['gen_ts']
            row.append(age(ct,ts))
        except:
            row.append('Unknown')
        # -o wide
        if output == 'wide':
            # containers
            conts = [ c['name'] for c in rs['spec']['template']['spec']['containers'] ]
            row.append(','.join(conts))
            # images
            images = [ c['image'] for c in rs['spec']['template']['spec']['containers'] ]
            row.append(','.join(images))
            # selector
            try:
                ml = ','.join(
                    [ k+'='+v 
                        for k,v in rs['spec']['selector']['matchLabels'].items()
                    ]
                )
            except:
                ml = '?'
            row.append(ml)

        output_res.append(row)

    print(tabulate(output_res,tablefmt="plain"))



from tabulate import tabulate

from omg.common.helper import age

def mc_out(t, ns, res, output, show_type):
    output_res=[]
    # header
    header = ['NAME','GENERATEDBYCONTROLLER','IGNITIONVERSION','AGE']
    # resources
    for r in res:
        mc = r['res']
        row = []
        # name
        if show_type:
            row.append(t + '/' + mc['metadata']['name'])
        else:
            row.append(mc['metadata']['name'])

        # generated by cont.
        try: 
            gen_by = mc['metadata']['annotations']['machineconfiguration.openshift.io/generated-by-controller-version']
        except:
            gen_by = ''
        row.append(gen_by)
        # ignition ver
        i_ver = mc['spec']['config']['ignition']['version']
        row.append(i_ver)
        # age
        try:
            ct = str(mc['metadata']['creationTimestamp'])
            ts = r['gen_ts']
            row.append(age(ct,ts))
        except:
            row.append('Unknown')

        output_res.append(row)

    #print(tabulate(output_res,tablefmt="plain"))
    # sort by 1st column (name)
    sorted_output = sorted(output_res, key=lambda x: x[0])
    sorted_output.insert(0,header)

    print(tabulate(sorted_output,tablefmt="plain"))



from tabulate import tabulate
from dateutil.parser import parse

from omg.common.helper import age

# Special function to output clusteroperator
def co_out(t, ns, res, output, show_type):
    # Generate output table if -o not set or 'wide'
    # We will create an array of array and then print if with tabulate
    # header
    output_res = [['NAME','VERSION','AVAILABLE','PROGRESSING','DEGRADED','SINCE']]
    # resources
    for r in res:
        co = r['res']
        row = []
        # name
        if show_type:
            row.append(t + '/' + co['metadata']['name'])
        else:
            row.append(co['metadata']['name'])
        # version
        vers = co['status']['versions']
        ver = next( v['version'] for v in vers if v['name'] == 'operator' )
        row.append(ver)
        # available,progressing,degraded,since
        transitions = []
        cond = co['status']['conditions']
        for c in cond:
            if c['type'] == 'Degraded':
                dg = c['status']
                transitions.append(str(c['lastTransitionTime']))
            elif c['type'] == 'Progressing':
                pr = c['status']
                transitions.append(str(c['lastTransitionTime']))
            elif c['type'] == 'Available':
                av = c['status']
                transitions.append(str(c['lastTransitionTime']))
        row.append(av)
        row.append(pr)
        row.append(dg)
        # since
        latest_trans = None
        for t in transitions:
            if latest_trans is None:
                latest_trans = t
            else:
                if parse(t) > parse(latest_trans):
                    latest_trans = t
        since = age(latest_trans,r['gen_ts'])
        row.append(since)

        output_res.append(row)

    print(tabulate(output_res,tablefmt="plain"))



import sys, os, yaml

CONFIG_FILE = os.getenv("HOME") + "/.omgconfig"

class Config:

    path = None
    project = None

    def __init__(self, fail_if_no_path=True):
        if os.path.exists(CONFIG_FILE):
            with open(CONFIG_FILE, 'r') as cf:
                c = yaml.safe_load(cf)
                if c is not None:
                    if 'path' in c:
                        Config.path = c['path']
                    if 'project' in c:
                        Config.project = c['project']
        if fail_if_no_path and Config.path is None:
            print('[ERROR] You have not selected a must-gather')
            print()
            print('Use `omg use </path/to/must-gather>` to point to an extracted must-gather.')
            sys.exit(1)

    def save(self, path=None, project=None):
        c = {}
        if path is not None:
            Config.path = path
            c['path'] = path
        else:
            c['path'] = Config.path

        if project is not None:
            Config.project = project
            c['project'] = project
        else:
            c['project'] = Config.project
        try:
            with open(CONFIG_FILE, 'w') as cf:
                yaml.dump(c, cf, default_flow_style=False)
        except IOError:
            print("[ERROR] Could not write config file:", CONFIG_FILE)



from tabulate import tabulate

from omg.common.helper import age

# Config Map
def cm_out(t, ns, res, output, show_type):
    output_res=[[]]
    # header
    if ns == '_all':
        output_res[0].append('NAMESPACE')
    output_res[0].extend(['NAME','DATA','AGE'])
    # resources
    for r in res:
        cm = r['res']
        row = []
        # namespace (for --all-namespaces)
        if ns == '_all':
            row.append(cm['metadata']['namespace'])
        # name
        if show_type:
            row.append(t + '/' + cm['metadata']['name'])
        else:
            row.append(cm['metadata']['name'])
        # data
        if 'data' in cm:
            row.append( len(cm['data']) )
        else:
            row.append('0')
        # age
        try:
            ct = str(cm['metadata']['creationTimestamp'])
            ts = r['gen_ts']
            row.append(age(ct,ts))
        except:
            row.append('Unknown')

        output_res.append(row)

    print(tabulate(output_res,tablefmt="plain"))



from tabulate import tabulate

from omg.common.helper import age

def mcp_out(t, ns, res, output, show_type):
    output_res=[[]]
    # header
    output_res[0].extend(['NAME','CONFIG','UPDATED','UPDATING','DEGRADED','MACHINECOUNT',
                          'READYMACHINECOUNT', 'UPDATEDMACHINECOUNT','DEGRADEDMACHINECOUNT','AGE'])
    # resources
    for r in res:
        mcp = r['res']
        row = []
        # name
        if show_type:
            row.append(t + '/' + mcp['metadata']['name'])
        else:
            row.append(mcp['metadata']['name'])
        # config
        row.append(mcp['spec']['configuration']['name'])
        # updated, updating, degraded
        updated = ''
        updating = ''
        degraded = ''
        conds = mcp['status']['conditions']
        for c in conds:
            if c['type'] == 'Updated':
                updated = c['status']
            if c['type'] == 'Updating':
                updating = c['status']
            if c['type'] == 'Degraded':
                degraded = c['status']
        row.append(updated)
        row.append(updating)
        row.append(degraded)
        # machine count
        row.append(mcp['status']['machineCount'])
        row.append(mcp['status']['readyMachineCount'])
        row.append(mcp['status']['updatedMachineCount'])
        row.append(mcp['status']['degradedMachineCount'])
        # age
        try:
            ct = str(mcp['metadata']['creationTimestamp'])
            ts = r['gen_ts']
            row.append(age(ct,ts))
        except:
            row.append('Unknown')

        output_res.append(row)

    print(tabulate(output_res,tablefmt="plain"))



from tabulate import tabulate
from dateutil.parser import parse

from omg.common.helper import age

# Simple out put with just name and age
def cv_out(t, ns, res, output, show_type):
    output_res=[[]]
    # header
    output_res[0].extend(['NAME','VERSION','AVAILABLE','PROGRESSING','SINCE','STATUS'])
    # resources
    for r in res:
        cv = r['res']
        row = []
        # name
        if show_type:
            row.append(t + '/' + cv['metadata']['name'])
        else:
            row.append(cv['metadata']['name'])
        # version
        cur_ver = ''
        history = cv['status']['history']
        for h in history:
            if h['state'] == 'Completed':
                cur_ver = h['version']
            break
        row.append(cur_ver)
        # available,progressing,status
        avail = ''
        prog = ''
        status = ''
        transitions = []
        conds = cv['status']['conditions']
        for c in conds:
            if c['type'] == 'Available':
                avail = c['status']
                transitions.append(str(c['lastTransitionTime']))
            elif c['type'] == 'Progressing':
                prog = c['status']
                status = c['message']
                transitions.append(str(c['lastTransitionTime']))
            elif c['type'] == 'Failing':
                transitions.append(str(c['lastTransitionTime']))
            
        row.append(avail)
        row.append(prog)
        # since
        latest_trans = None
        for t in transitions:
            if latest_trans is None:
                latest_trans = t
            else:
                if parse(t) > parse(latest_trans):
                    latest_trans = t
        since = age(latest_trans,r['gen_ts'])
        row.append(since)

        # status
        row.append(status)

        output_res.append(row)

    print(tabulate(output_res,tablefmt="plain"))



from tabulate import tabulate

from omg.common.helper import age


def pv_out(t, ns, res, output, show_type):
    output_res=[[]]
    # header
    output_res[0].extend(['NAME','CAPACITY','ACCESS MODES','RECLAIM POLICY',
                            'STATUS','CLAIM','STORAGECLASS','REASON','AGE'])
    # resources
    for r in res:
        p = r['res']
        row = []
        # name
        if show_type:
            row.append(t + '/' + p['metadata']['name'])
        else:
            row.append(p['metadata']['name'])
        # capacity
        row.append(p['spec']['capacity']['storage'])
        # access
        access = ",".join(p['spec']['accessModes'])
        row.append(access)
        # reclaim
        row.append(p['spec']['persistentVolumeReclaimPolicy'])
        # status
        row.append(p['status']['phase'])
        # claim
        try:
            cr_name = p['spec']['claimRef']['name']
            cr_ns = p['spec']['claimRef']['namespace']
            claim = cr_ns + '/' + cr_name
        except:
            claim = ''
        row.append(claim)
        # storage class
        if 'storageClassName' in p['spec']:
            row.append(p['spec']['storageClassName'])
        else:
            row.append('')
        
        # reason (?) TODO need sample of this yaml
        row.append('')
        # age
        try:
            ct = p['metadata']['creationTimestamp']
            ts = r['gen_ts']
            row.append(age(ct,ts))
        except:
            row.append('Unknown')

        output_res.append(row)

    print(tabulate(output_res,tablefmt="plain"))



from tabulate import tabulate

from omg.common.helper import age


# Simple out put with just name and age
def route_out(t, ns, res, output, show_type):
    output_res=[[]]
    # header
    if ns == '_all':
        output_res[0].append('NAMESPACE')
    output_res[0].extend(['NAME','HOST/PORT','PATH','SERVICES','PORT','TERMINATION','WILDCARD'])
    # resources
    for r in res:
        route = r['res']
        row = []
        # namespace (for --all-namespaces)
        if ns == '_all':
            row.append(route['metadata']['namespace'])
        # name
        if show_type:
            row.append(t + '/' + route['metadata']['name'])
        else:
            row.append(route['metadata']['name'])
        # host/port
        row.append(route['spec']['host'])
        # path
        if 'path' in route['spec']:
            row.append(route['spec']['path'])
        else:
            row.append('')
        # services
        try:
            svc = route['spec']['to']['name'] if route['spec']['to']['kind'] == 'Service' else '?'
            row.append(svc)
        except:
            row.append('??')
        # port
        try:
            port = route['spec']['port']['targetPort']
            row.append(port)
        except:
            row.append('??')
        # termination
        try:
            term = route['spec']['tls']
            term_str = ''
            if 'termination' in term:
                term_str += term['termination']
            if 'insecureEdgeTerminationPolicy' in term:
                term_str += '/' + term['insecureEdgeTerminationPolicy']
            row.append(term_str)
        except:
            row.append('??')
        # wildcard
        try:
            row.append(route['spec']['wildcardPolicy'])
        except:
            row.append('??')
        # # age
        # try:
        #     ct = p['metadata']['creationTimestamp']
        #     ts = r['gen_ts']
        #     row.append(age(ct,ts))
        # except:
        #     row.append('Unknown')

        output_res.append(row)

    print(tabulate(output_res,tablefmt="plain"))



from tabulate import tabulate

from omg.common.helper import age


# Simple out put with just name and age
def deployment_out(t, ns, res, output, show_type):
    output_res=[[]]
    # header
    if ns == '_all':
        output_res[0].append('NAMESPACE')
    output_res[0].extend(['NAME','READY','UP-TO-DATE','AVAILABLE','AGE'])
    if output == 'wide':
        output_res[0].extend(['CONTAINERS','IMAGES'])
    # resources
    for r in res:
        dep = r['res']
        row = []
        # namespace (for --all-namespaces)
        if ns == '_all':
            row.append(dep['metadata']['namespace'])
        # name
        if show_type:
            row.append(t + '/' + dep['metadata']['name'])
        else:
            row.append(dep['metadata']['name'])
        # ready
        try:
            ready = dep['status']['readyReplicas']
            replicas = dep['status']['replicas']
            row.append( str(ready) + '/' + str(replicas) )
        except:
            row.append('??')
        # up-to-date
        try:
            row.append(dep['status']['updatedReplicas'])
        except:
            row.append('??')
        # available
        try:
            row.append(dep['status']['availableReplicas'])
        except:
            row.append('??')
        # age
        try:
            ct = str(dep['metadata']['creationTimestamp'])
            ts = r['gen_ts']
            row.append(age(ct,ts))
        except:
            row.append('Unknown')
        # -o wide
        if output == 'wide':
            # container:
            conts = [ c['name'] for c in dep['spec']['template']['spec']['containers'] ]
            row.append(','.join(conts))
            # images:
            images = [ c['image'] for c in dep['spec']['template']['spec']['containers'] ]
            row.append(','.join(images))

        output_res.append(row)

    print(tabulate(output_res,tablefmt="plain"))



from omg.cmd.get.from_yaml import from_yaml
from omg.cmd.get.simple_out import simple_out
from omg.cmd.get.node_out import node_out
from omg.cmd.get.pod_out import pod_out
from omg.cmd.get.co_out import co_out
from omg.cmd.get.cv_out import cv_out
from omg.cmd.get.mcp_out import mcp_out
from omg.cmd.get.mc_out import mc_out
from omg.cmd.get.pv_out import pv_out
from omg.cmd.get.secret_out import secret_out
from omg.cmd.get.service_out import service_out
from omg.cmd.get.route_out import route_out
from omg.cmd.get.cm_out import cm_out
from omg.cmd.get.ep_out import ep_out
from omg.cmd.get.ev_out import ev_out
from omg.cmd.get.deployment_out import deployment_out
from omg.cmd.get.rs_out import rs_out
from omg.cmd.get.ss_out import ss_out


# This map and function normalizes/standarizes the different names of object/resources types
# Also returns the function that handles get,describe and output of this object type
# This is how the top level get_main function knows which objects we can handle and how.
# returns None if not an object type that we can't handle


map = [
    {   'type': 'pod','aliases': ['pods', 'po'],'need_ns': True,
        'get_func': from_yaml, 'getout_func': pod_out,
        'yaml_loc': 'namespaces/%s/core/pods.yaml' },

    {   'type': 'service','aliases': ['services', 'svc'],'need_ns': True,
        'get_func': from_yaml, 'getout_func': service_out,
        'yaml_loc': 'namespaces/%s/core/services.yaml' },

    {   'type': 'secret','aliases': ['secrets'],'need_ns': True,
        'get_func': from_yaml, 'getout_func': secret_out,
        'yaml_loc': 'namespaces/%s/core/secrets.yaml' },

    {   'type': 'configmap','aliases': ['configmaps', 'cm'],'need_ns': True,
        'get_func': from_yaml, 'getout_func': cm_out,
        'yaml_loc': 'namespaces/%s/core/configmaps.yaml' },

    {   'type': 'endpoint','aliases': ['endpoints', 'ep'],'need_ns': True,
        'get_func': from_yaml, 'getout_func': ep_out,
        'yaml_loc': 'namespaces/%s/core/endpoints.yaml' },

    {   'type': 'event','aliases': ['events', 'ev'],'need_ns': True,
        'get_func': from_yaml, 'getout_func': ev_out,
        'yaml_loc': 'namespaces/%s/core/events.yaml' },

    {   'type': 'persistentvolumeclaim','aliases': ['persistentvolumeclaims', 'pvc'],'need_ns': True,
        'get_func': from_yaml, 'getout_func': simple_out,
        'yaml_loc': 'namespaces/%s/core/persistentvolumeclaims.yaml' }, ###

    {   'type': 'replicationcontroller','aliases': ['replicationcontrollers', 'rc'],'need_ns': True,
        'get_func': from_yaml, 'getout_func': simple_out,
        'yaml_loc': 'namespaces/%s/core/replicationcontrollers.yaml' }, ###

    {   'type': 'replicaset','aliases': ['replicasets', 'rs'],'need_ns': True,
        'get_func': from_yaml, 'getout_func': rs_out,
        'yaml_loc': 'namespaces/%s/apps/replicasets.yaml' },

    {   'type': 'statefulset','aliases': ['statefulsets'],'need_ns': True,
        'get_func': from_yaml, 'getout_func': ss_out,
        'yaml_loc': 'namespaces/%s/apps/statefulsets.yaml' }, ###

    {   'type': 'deployment','aliases': ['deployments'],'need_ns': True,
        'get_func': from_yaml, 'getout_func': deployment_out,
        'yaml_loc': 'namespaces/%s/apps/deployments.yaml' },
    
    {   'type': 'daemonset','aliases': ['daemonsets', 'ds'],'need_ns': True,
        'get_func': from_yaml, 'getout_func': simple_out,
        'yaml_loc': 'namespaces/%s/apps/daemonsets.yaml' }, ###

    {   'type': 'deploymentconfig','aliases': ['deploymentconfigs', 'dc'],'need_ns': True,
        'get_func': from_yaml, 'getout_func': simple_out,
        'yaml_loc': 'namespaces/%s/apps.openshift.io/deploymentconfigs.yaml' }, ###

    {   'type': 'route','aliases': ['routes'],'need_ns': True,
        'get_func': from_yaml, 'getout_func': route_out,
        'yaml_loc': 'namespaces/%s/route.openshift.io/routes.yaml' }, ###

    {   'type': 'node','aliases': ['nodes'],'need_ns': False,
        'get_func': from_yaml, 'getout_func': node_out,
        'yaml_loc': 'cluster-scoped-resources/core/nodes' },
        # When yaml_loc is a dir like in this case, we scan *.yaml files in it.

    {   'type': 'persistentvolume','aliases': ['persistentvolumes', 'pv'],'need_ns': False,
        'get_func': from_yaml, 'getout_func': pv_out,
        'yaml_loc': 'cluster-scoped-resources/core/persistentvolumes' },

    {   'type': 'machineconfigpool','aliases': ['machineconfigpools', 'mcp'],'need_ns': False,
        'get_func': from_yaml, 'getout_func': mcp_out,
        'yaml_loc': 'cluster-scoped-resources/machineconfiguration.openshift.io/machineconfigpools' },

    {   'type': 'machineconfig','aliases': ['machineconfigs', 'mc'],'need_ns': False,
        'get_func': from_yaml, 'getout_func': mc_out,
        'yaml_loc': 'cluster-scoped-resources/machineconfiguration.openshift.io/machineconfigs' },

    {   'type': 'clusteroperator','aliases': ['clusteroperators', 'co'],'need_ns': False,
        'get_func': from_yaml, 'getout_func': co_out,
        'yaml_loc': 'cluster-scoped-resources/config.openshift.io/clusteroperators.yaml' },

    {   'type': 'clusterversion','aliases': [],'need_ns': False,
        'get_func': from_yaml, 'getout_func': cv_out,
        'yaml_loc': 'cluster-scoped-resources/config.openshift.io/clusterversions.yaml' },

    {   'type': 'console','aliases': ['consoles'],'need_ns': False,
        'get_func': from_yaml, 'getout_func': simple_out,
        'yaml_loc': 'cluster-scoped-resources/config.openshift.io/consoles.yaml' },

    {   'type': 'dns','aliases': ['dnses'],'need_ns': False,
        'get_func': from_yaml, 'getout_func': simple_out,
        'yaml_loc': 'cluster-scoped-resources/config.openshift.io/dnses.yaml' },

    {   'type': 'featuregate','aliases': ['featuregates'],'need_ns': False,
        'get_func': from_yaml, 'getout_func': simple_out,
        'yaml_loc': 'cluster-scoped-resources/config.openshift.io/featuregates.yaml' },

    {   'type': 'image','aliases': ['images'],'need_ns': False,
        'get_func': from_yaml, 'getout_func': simple_out,
        'yaml_loc': 'cluster-scoped-resources/config.openshift.io/images.yaml' },

    {   'type': 'infrastructure','aliases': ['infrastructures'],'need_ns': False,
        'get_func': from_yaml, 'getout_func': simple_out,
        'yaml_loc': 'cluster-scoped-resources/config.openshift.io/infrastructures.yaml' },

    {   'type': 'ingress','aliases': ['ingresses'],'need_ns': False,
        'get_func': from_yaml, 'getout_func': simple_out,
        'yaml_loc': 'cluster-scoped-resources/config.openshift.io/ingresses.yaml' },

    {   'type': 'network','aliases': ['networks'],'need_ns': False,
        'get_func': from_yaml, 'getout_func': simple_out,
        'yaml_loc': 'cluster-scoped-resources/config.openshift.io/networks.yaml' },

    {   'type': 'oauth','aliases': ['oauths'],'need_ns': False,
        'get_func': from_yaml, 'getout_func': simple_out,
        'yaml_loc': 'cluster-scoped-resources/config.openshift.io/oauths.yaml' },

    {   'type': 'operatorhub','aliases': ['operatorhubs'],'need_ns': False,
        'get_func': from_yaml, 'getout_func': simple_out,
        'yaml_loc': 'cluster-scoped-resources/config.openshift.io/operatorhubs.yaml' },

    {   'type': 'project','aliases': ['projects'],'need_ns': False,
        'get_func': from_yaml, 'getout_func': simple_out,
        'yaml_loc': 'cluster-scoped-resources/config.openshift.io/projects.yaml' },

    {   'type': 'proxy','aliases': ['proxies'],'need_ns': False,
        'get_func': from_yaml, 'getout_func': simple_out,
        'yaml_loc': 'cluster-scoped-resources/config.openshift.io/proxies.yaml' },

    {   'type': 'scheduler','aliases': ['schedulers'],'need_ns': False,
        'get_func': from_yaml, 'getout_func': simple_out,
        'yaml_loc': 'cluster-scoped-resources/config.openshift.io/schedulers.yaml' },
]


def map_res(t):
    if t is not None:
        for x in map:
            # match the input with type: or alias (without case)
            if t.lower() == x['type'].lower() or t.lower() in [y.lower() for y in x['aliases'] ]:
                return x
    return None
